#define _GNU_SOURCE
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <search.h>

#include "list.h"

#define NPKG 5
#define MINC 128

struct call {
  unsigned long nr;
  char *name, pkg;
};

static struct node pq[NPKG];
static void * nameroot;
static void * nrroot;

static void
free_call(void * c) {
  free(((struct call *) c)->name);
  free(c);
}

static void do_nothing(void * n) { }

static char *
readname()
{
  size_t size = MINC, i = 0;
  char * name = malloc(sizeof(char)*MINC), c;

  while ((c = getchar()) != ' ' && c != '\n') {
    if (i >= size - 1) {
      size += MINC;
      name = realloc(name, sizeof(char)*size);
    }
    name[i++] = c;
  }
  name[i] = '\0';

  return name;
}

static int
nrcmp (const void *pa, const void *pb)
{
  struct call * la = (struct call *) ((link) pa)->value;
  struct call * lb = (struct call *) ((link) pb)->value;

  if (la->nr < lb->nr)
    return -1;
  if (la->nr > lb->nr)
    return 1;
  return 0;
}

static int
namecmp(const void *pa, const void *pb)
{
  struct call * la = (struct call *) ((link) pa)->value;
  struct call * lb = (struct call *) ((link) pb)->value;

  return strcmp(la->name, lb->name);
}

int
main(int argc, char * argv[])
{
  char cmd;
  struct call * c, ctmp;
  struct node ntmp;
  link ltmp;
  size_t i;

  memset(pq, 0x0, sizeof(struct node)*NPKG);
  for (i = 0; i < NPKG; i++)
    pq[i].next = pq[i].prev = &pq[i];
  nameroot = nrroot = NULL;

  while ((cmd = getchar()) != 'x') {
    switch (cmd) {
      case 'a':
        c = malloc(sizeof(struct call));
        getchar(); /* catch ' ' */
        c->name = readname();
        scanf("%lu", &c->nr);
        scanf(" %c", &c->pkg);
        getchar(); /* catch '\n' */
        list_insert(&pq[c->pkg - 'A'], c);
        tsearch(pq[c->pkg - 'A'].prev, &nameroot, namecmp);
        tsearch(pq[c->pkg - 'A'].prev, &nrroot, nrcmp);
        break;
      case 'u':
        scanf("%lu", &ctmp.nr);
        scanf(" %c", &ctmp.pkg);
        getchar(); /* catch '\n' */
        ntmp.value = &ctmp;
        ltmp = *((link*) tfind(&ntmp, &nrroot, nrcmp));
        if (ctmp.pkg != ((struct call *) ltmp->value)->pkg) {
          c = (struct call *) ltmp->value;
          ((struct call *) ltmp->value)->pkg = ctmp.pkg;
          list_move(&pq[c->pkg - 'A'], ltmp);
        }
        break;
      case 'l':
        getchar(); /* catch '\n' */
        for (i = 0; i < NPKG && pq[i].next == &pq[i]; i++);
        if (i < NPKG) {
          c = (struct call *) pq[i].next->value;
          printf("%s %lu %c\n", c->name, c->nr, c->pkg);
        } else
          printf("NA\n");
        break;
      case 'p':
        getchar(); /* catch '\n' */
        for (i = 0; i < NPKG && pq[i].next == &pq[i]; i++);
        if (i < NPKG) {
          c = (struct call *) pq[i].next->value;
          printf("%s %lu %c\n", c->name, c->nr, c->pkg);
          tdelete(pq[i].next, &nameroot, namecmp);
          tdelete(pq[i].next, &nrroot, nrcmp);
          list_delete(pq[i].next);
          free_call(c);
        } else
          printf("NA\n");
        break;
      case 'r':
        getchar(); /* catch ' ' */
        ctmp.name = readname();
        ntmp.value = &ctmp;
        /*fprintf(stderr, "ERROR: %s\n", ctmp.name);*/
        ltmp = *((link *) tfind(&ntmp, &nameroot, namecmp));
        c = (struct call *) ltmp->value;
        tdelete(ltmp, &nameroot, namecmp);
        tdelete(ltmp, &nrroot, nrcmp);
        list_delete(ltmp);
        free_call(c);
        free(ctmp.name);
    }
  }

  tdestroy(nameroot, do_nothing);
  tdestroy(nrroot, do_nothing);
  for (i = 0; i < NPKG; i++)
    list_free(&pq[i], free_call);

  return EXIT_SUCCESS;
}

